@@ -30,7 +30,7 @@
 #include <time.h>
 #include <stdio.h>
 #include <ctype.h>
-#include <utmp.h>
+#include <utmpx.h>
 #include <pwd.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -40,6 +40,7 @@
 #include <netinet/in.h>
 #include <netdb.h>
 #include <arpa/inet.h>
+#include <libgen.h>
 
 #include "c.h"
 #include "nls.h"
@@ -51,11 +52,10 @@
 #include "strutils.h"
 #include "timeutils.h"
 #include "monotonic.h"
+#include "fileutils.h"
 
-#if defined(_HAVE_UT_TV)
-# define UL_UT_TIME ut_tv.tv_sec
-#else
-# define UL_UT_TIME ut_time
+#ifdef FUZZ_TARGET
+#include "fuzz.h"
 #endif
 
 #ifndef SHUTDOWN_TIME
@@ -95,11 +95,12 @@ struct last_control {
 	time_t until;		/* at what time to stop displaying the file */
 	time_t present;		/* who where present at time_t */
 	unsigned int time_fmt;	/* time format */
+	char separator;        /* output separator */
 };
 
 /* Double linked list of struct utmp's */
 struct utmplist {
-	struct utmp ut;
+	struct utmpx ut;
 	struct utmplist *next;
 	struct utmplist *prev;
 };
@@ -111,28 +112,51 @@ enum {
 	R_NORMAL,	/* Normal */
 	R_NOW,		/* Still logged in */
 	R_REBOOT,	/* Reboot record. */
+	R_REBOOT_CRASH,	/* Reboot record without matching shutdown */
 	R_PHANTOM,	/* No logout record but session is stale. */
 	R_TIMECHANGE	/* NEW_TIME or OLD_TIME */
 };
 
 enum {
 	LAST_TIMEFTM_NONE = 0,
-	LAST_TIMEFTM_SHORT_CTIME,
-	LAST_TIMEFTM_FULL_CTIME,
-	LAST_TIMEFTM_ISO8601
+	LAST_TIMEFTM_SHORT,
+	LAST_TIMEFTM_CTIME,
+	LAST_TIMEFTM_ISO8601,
+
+	LAST_TIMEFTM_HHMM,	/* non-public */
 };
 
 struct last_timefmt {
 	const char *name;
-	int in;
-	int out;
+	int in_len;	/* log-in */
+	int in_fmt;
+	int out_len;	/* log-out */
+	int out_fmt;
 };
 
 static struct last_timefmt timefmts[] = {
-	[LAST_TIMEFTM_NONE]	   = { "notime", 0, 0 },
-	[LAST_TIMEFTM_SHORT_CTIME] = { "short", 16, 7},
-	[LAST_TIMEFTM_FULL_CTIME]  = { "full",  24, 26},
-	[LAST_TIMEFTM_ISO8601]	   = { "iso", 24, 26}
+	[LAST_TIMEFTM_NONE] = { .name = "notime" },
+	[LAST_TIMEFTM_SHORT] = {
+		.name    = "short",
+		.in_len  = 16,
+		.out_len = 7,
+		.in_fmt  = LAST_TIMEFTM_CTIME,
+		.out_fmt = LAST_TIMEFTM_HHMM
+	},
+	[LAST_TIMEFTM_CTIME] = {
+		.name    = "full",
+		.in_len  = 24,
+		.out_len = 26,
+		.in_fmt  = LAST_TIMEFTM_CTIME,
+		.out_fmt = LAST_TIMEFTM_CTIME
+	},
+	[LAST_TIMEFTM_ISO8601] = {
+		.name    = "iso",
+		.in_len  = 25,
+		.out_len = 27,
+		.in_fmt  = LAST_TIMEFTM_ISO8601,
+		.out_fmt = LAST_TIMEFTM_ISO8601
+	}
 };
 
 /* Global variables */
@@ -140,23 +164,25 @@ static unsigned int recsdone;	/* Number of records listed */
 static time_t lastdate;		/* Last date we've seen */
 static time_t currentdate;	/* date when we started processing the file */
 
+#ifndef FUZZ_TARGET
 /* --time-format=option parser */
-static int which_time_format(const char *optarg)
+static int which_time_format(const char *s)
 {
 	size_t i;
 
 	for (i = 0; i < ARRAY_SIZE(timefmts); i++) {
-		if (strcmp(timefmts[i].name, optarg) == 0)
+		if (strcmp(timefmts[i].name, s) == 0)
 			return i;
 	}
-	errx(EXIT_FAILURE, _("unknown time format: %s"), optarg);
+	errx(EXIT_FAILURE, _("unknown time format: %s"), s);
 }
+#endif
 
 /*
  *	Read one utmp entry, return in new format.
  *	Automatically reposition file pointer.
  */
-static int uread(FILE *fp, struct utmp *u,  int *quit, const char *filename)
+static int uread(FILE *fp, struct utmpx *u,  int *quit, const char *filename)
 {
 	static int utsize;
 	static char buf[UCHUNKSIZE];
@@ -169,14 +195,14 @@ static int uread(FILE *fp, struct utmp *u,  int *quit, const char *filename)
 		/*
 		 *	Normal read.
 		 */
-		return fread(u, sizeof(struct utmp), 1, fp);
+		return fread(u, sizeof(struct utmpx), 1, fp);
 	}
 
 	if (u == NULL) {
 		/*
 		 *	Initialize and position.
 		 */
-		utsize = sizeof(struct utmp);
+		utsize = sizeof(struct utmpx);
 		fseeko(fp, 0, SEEK_END);
 		fpos = ftello(fp);
 		if (fpos == 0)
@@ -200,7 +226,7 @@ static int uread(FILE *fp, struct utmp *u,  int *quit, const char *filename)
 	 */
 	bpos -= utsize;
 	if (bpos >= 0) {
-		memcpy(u, buf + bpos, sizeof(struct utmp));
+		memcpy(u, buf + bpos, sizeof(struct utmpx));
 		return 1;
 	}
 
@@ -236,27 +262,18 @@ static int uread(FILE *fp, struct utmp *u,  int *quit, const char *filename)
 	memcpy(tmp, buf + UCHUNKSIZE + bpos, -bpos);
 	bpos += UCHUNKSIZE;
 
-	memcpy(u, tmp, sizeof(struct utmp));
+	memcpy(u, tmp, sizeof(struct utmpx));
 
 	return 1;
 }
 
-/*
- *	Print a short date.
- */
-static char *showdate(void)
-{
-	char *s = ctime(&lastdate);
-	s[16] = 0;
-	return s;
-}
-
+#ifndef FUZZ_TARGET
 /*
  *	SIGINT handler
  */
 static void int_handler(int sig __attribute__((unused)))
 {
-	errx(EXIT_FAILURE, _("Interrupted %s"), showdate());
+	ul_sig_err(EXIT_FAILURE, "Interrupted");
 }
 
 /*
@@ -264,9 +281,10 @@ static void int_handler(int sig __attribute__((unused)))
  */
 static void quit_handler(int sig __attribute__((unused)))
 {
-	warnx(_("Interrupted %s"), showdate());
+	ul_sig_warn("Interrupted");
 	signal(SIGQUIT, quit_handler);
 }
+#endif
 
 /*
  *	Lookup a host with DNS.
@@ -312,39 +330,38 @@ static int dns_lookup(char *result, int size, int useip, int32_t *a)
 	return getnameinfo(sa, salen, result, size, NULL, 0, flags);
 }
 
-static int time_formatter(const struct last_control *ctl, char *dst,
-			  size_t dlen, time_t *when, int pos)
+static int time_formatter(int fmt, char *dst, size_t dlen, time_t *when)
 {
-	struct tm *tm;
 	int ret = 0;
 
-	switch (ctl->time_fmt) {
+	switch (fmt) {
 	case LAST_TIMEFTM_NONE:
 		*dst = 0;
 		break;
-	case LAST_TIMEFTM_SHORT_CTIME:
-		if (pos == 0)
-			ret = sprintf(dst, "%s", ctime(when));
-		else {
-			tm = localtime(when);
-			if (!strftime(dst, dlen, "- %H:%M", tm))
-				ret = -1;
-		}
-		break;
-	case LAST_TIMEFTM_FULL_CTIME:
-		if (pos == 0)
-			ret = sprintf(dst, "%s", ctime(when));
-		else
-			ret = sprintf(dst, "- %s", ctime(when));
-		break;
-	case LAST_TIMEFTM_ISO8601:
-		tm = localtime(when);
-		if (pos == 0) {
-			if (!strftime(dst, dlen, "%Y-%m-%dT%H:%M:%S%z", tm))
-				ret = -1;
-		} else if (!strftime(dst, dlen, "- %Y-%m-%dT%H:%M:%S%z", tm))
+	case LAST_TIMEFTM_HHMM:
+	{
+		struct tm tm;
+
+		localtime_r(when, &tm);
+		if (!snprintf(dst, dlen, "%02d:%02d", tm.tm_hour, tm.tm_min))
 			ret = -1;
 		break;
+	}
+	case LAST_TIMEFTM_CTIME:
+	{
+		char buf[CTIME_BUFSIZ];
+
+		if (!ctime_r(when, buf)) {
+			ret = -1;
+			break;
+		}
+		snprintf(dst, dlen, "%s", buf);
+		ret = rtrim_whitespace((unsigned char *) dst);
+		break;
+	}
+	case LAST_TIMEFTM_ISO8601:
+		ret = strtime_iso(when, ISO_TIMESTAMP_T, dst, dlen);
+		break;
 	default:
 		abort();
 	}
@@ -371,16 +388,15 @@ static void trim_trailing_spaces(char *s)
 /*
  *	Show one line of information on screen
  */
-static int list(const struct last_control *ctl, struct utmp *p, time_t logout_time, int what)
+static int list(const struct last_control *ctl, struct utmpx *p, time_t logout_time, int what)
 {
 	time_t		secs, utmp_time;
 	char		logintime[LAST_TIMESTAMP_LEN];
 	char		logouttime[LAST_TIMESTAMP_LEN];
 	char		length[LAST_TIMESTAMP_LEN];
 	char		final[512];
-	char		utline[UT_LINESIZE+1];
+	char		utline[sizeof(p->ut_line) + 1];
 	char		domain[256];
-	char		*s;
 	int		mins, hours, days;
 	int		r, len;
 	struct last_timefmt *fmt;
@@ -388,8 +404,7 @@ static int list(const struct last_control *ctl, struct utmp *p, time_t logout_ti
 	/*
 	 *	uucp and ftp have special-type entries
 	 */
-	utline[0] = 0;
-	strncat(utline, p->ut_line, UT_LINESIZE);
+	mem2strcpy(utline, p->ut_line, sizeof(p->ut_line), sizeof(utline));
 	if (strncmp(utline, "ftp", 3) == 0 && isdigit(utline[3]))
 		utline[3] = 0;
 	if (strncmp(utline, "uucp", 4) == 0 && isdigit(utline[4]))
@@ -401,7 +416,7 @@ static int list(const struct last_control *ctl, struct utmp *p, time_t logout_ti
 	if (ctl->show) {
 		char **walk;
 		for (walk = ctl->show; *walk; walk++) {
-			if (strncmp(p->ut_user, *walk, UT_NAMESIZE) == 0 ||
+			if (strncmp(p->ut_user, *walk, sizeof(p->ut_user)) == 0 ||
 			    strcmp(utline, *walk) == 0 ||
 			    (strncmp(utline, "tty", 3) == 0 &&
 			     strcmp(utline + 3, *walk) == 0)) break;
@@ -412,7 +427,9 @@ static int list(const struct last_control *ctl, struct utmp *p, time_t logout_ti
 	/*
 	 *	Calculate times
 	 */
-	utmp_time = p->UL_UT_TIME;
+	fmt = &timefmts[ctl->time_fmt];
+
+	utmp_time = p->ut_tv.tv_sec;
 
 	if (ctl->present) {
 		if (ctl->present < utmp_time)
@@ -420,63 +437,76 @@ static int list(const struct last_control *ctl, struct utmp *p, time_t logout_ti
 		if (0 < logout_time && logout_time < ctl->present)
 			return 0;
 	}
-	if (time_formatter(ctl, &logintime[0], sizeof(logintime), &utmp_time, 0) < 0 ||
-	    time_formatter(ctl, &logouttime[0], sizeof(logouttime), &logout_time, 1) < 0)
+
+	/* log-in time */
+	if (time_formatter(fmt->in_fmt, logintime,
+			   sizeof(logintime), &utmp_time) < 0)
 		errx(EXIT_FAILURE, _("preallocation size exceeded"));
 
-	secs  = logout_time - utmp_time;
+	/* log-out time */
+	secs  = logout_time - utmp_time; /* Under strange circumstances, secs < 0 can happen */
 	mins  = (secs / 60) % 60;
 	hours = (secs / 3600) % 24;
 	days  = secs / 86400;
 
+	strcpy(logouttime, "- ");
+	if (time_formatter(fmt->out_fmt, logouttime + 2,
+			   sizeof(logouttime) - 2, &logout_time) < 0)
+		errx(EXIT_FAILURE, _("preallocation size exceeded"));
+
 	if (logout_time == currentdate) {
-		if (ctl->time_fmt > LAST_TIMEFTM_SHORT_CTIME) {
-			sprintf(logouttime, "  still running");
+		if (ctl->time_fmt > LAST_TIMEFTM_SHORT) {
+			snprintf(logouttime, sizeof(logouttime), "  still running");
 			length[0] = 0;
 		} else {
-			sprintf(logouttime, "  still");
-			sprintf(length, "running");
+			snprintf(logouttime, sizeof(logouttime), "  still");
+			snprintf(length, sizeof(length), "running");
 		}
-	} else if (days)
-		sprintf(length, "(%d+%02d:%02d)", days, hours, mins);
-	else
-		sprintf(length, " (%02d:%02d)", hours, mins);
+	} else if (days) {
+		snprintf(length, sizeof(length), "(%d+%02d:%02d)", days, abs(hours), abs(mins)); /* hours and mins always shown as positive (w/o minus sign!) even if secs < 0 */
+	} else if (hours) {
+		snprintf(length, sizeof(length), " (%02d:%02d)", hours, abs(mins));  /* mins always shown as positive (w/o minus sign!) even if secs < 0 */
+	} else if (secs >= 0) {
+		snprintf(length, sizeof(length), " (%02d:%02d)", hours, mins);
+	} else {
+		snprintf(length, sizeof(length), " (-00:%02d)", abs(mins));  /* mins always shown as positive (w/o minus sign!) even if secs < 0 */
+	}
 
 	switch(what) {
 		case R_CRASH:
-			sprintf(logouttime, "- crash");
+		case R_REBOOT_CRASH:
+			snprintf(logouttime, sizeof(logouttime), "- crash");
 			break;
 		case R_DOWN:
-			sprintf(logouttime, "- down ");
+			snprintf(logouttime, sizeof(logouttime), "- down ");
 			break;
 		case R_NOW:
-			if (ctl->time_fmt > LAST_TIMEFTM_SHORT_CTIME) {
-				sprintf(logouttime, "  still logged in");
+			if (ctl->time_fmt > LAST_TIMEFTM_SHORT) {
+				snprintf(logouttime, sizeof(logouttime), "  still logged in");
 				length[0] = 0;
 			} else {
-				sprintf(logouttime, "  still");
-				sprintf(length, "logged in");
+				snprintf(logouttime, sizeof(logouttime), "  still");
+				snprintf(length, sizeof(length), "logged in");
 			}
 			break;
 		case R_PHANTOM:
-			if (ctl->time_fmt > LAST_TIMEFTM_SHORT_CTIME) {
-				sprintf(logouttime, "  gone - no logout");
+			if (ctl->time_fmt > LAST_TIMEFTM_SHORT) {
+				snprintf(logouttime, sizeof(logouttime), "  gone - no logout");
 				length[0] = 0;
-			} else if (ctl->time_fmt == LAST_TIMEFTM_SHORT_CTIME) {
-				sprintf(logouttime, "   gone");
-				sprintf(length, "- no logout");
+			} else if (ctl->time_fmt == LAST_TIMEFTM_SHORT) {
+				snprintf(logouttime, sizeof(logouttime), "   gone");
+				snprintf(length, sizeof(length), "- no logout");
 			} else {
 				logouttime[0] = 0;
-				sprintf(length, "no logout");
+				snprintf(length, sizeof(length), "no logout");
 			}
 			break;
-		case R_REBOOT:
-			break;
 		case R_TIMECHANGE:
 			logouttime[0] = 0;
 			length[0] = 0;
 			break;
 		case R_NORMAL:
+		case R_REBOOT:
 			break;
 		default:
 			abort();
@@ -487,37 +517,31 @@ static int list(const struct last_control *ctl, struct utmp *p, time_t logout_ti
 	 */
 	r = -1;
 	if (ctl->usedns || ctl->useip)
-		r = dns_lookup(domain, sizeof(domain), ctl->useip, p->ut_addr_v6);
-	if (r < 0) {
-		len = UT_HOSTSIZE;
-		if (len >= (int)sizeof(domain)) len = sizeof(domain) - 1;
-		domain[0] = 0;
-		strncat(domain, p->ut_host, len);
-	}
-
-	fmt = &timefmts[ctl->time_fmt];
+		r = dns_lookup(domain, sizeof(domain), ctl->useip, (int32_t*)p->ut_addr_v6);
+	if (r < 0)
+		mem2strcpy(domain, p->ut_host, sizeof(p->ut_host), sizeof(domain));
 
 	if (ctl->showhost) {
 		if (!ctl->altlist) {
 			len = snprintf(final, sizeof(final),
-				"%-8.*s %-12.12s %-16.*s %-*.*s %-*.*s %s\n",
-				ctl->name_len, p->ut_user, utline,
-				ctl->domain_len, domain,
-				fmt->in, fmt->in, logintime, fmt->out, fmt->out,
-				logouttime, length);
+				"%-8.*s%c%-12.12s%c%-16.*s%c%-*.*s%c%-*.*s%c%s\n",
+				ctl->name_len, p->ut_user, ctl->separator, utline, ctl->separator,
+				ctl->domain_len, domain, ctl->separator,
+				fmt->in_len, fmt->in_len, logintime, ctl->separator, fmt->out_len, fmt->out_len,
+				logouttime, ctl->separator, length);
 		} else {
 			len = snprintf(final, sizeof(final),
-				"%-8.*s %-12.12s %-*.*s %-*.*s %-12.12s %s\n",
-				ctl->name_len, p->ut_user, utline,
-				fmt->in, fmt->in, logintime, fmt->out, fmt->out,
-				logouttime, length, domain);
+				"%-8.*s%c%-12.12s%c%-*.*s%c%-*.*s%c%-12.12s%c%s\n",
+				ctl->name_len, p->ut_user, ctl->separator, utline, ctl->separator,
+				fmt->in_len, fmt->in_len, logintime, ctl->separator, fmt->out_len, fmt->out_len,
+				logouttime, ctl->separator, length, ctl->separator, domain);
 		}
 	} else
 		len = snprintf(final, sizeof(final),
-			"%-8.*s %-12.12s %-*.*s %-*.*s %s\n",
-			ctl->name_len, p->ut_user, utline,
-			fmt->in, fmt->in, logintime, fmt->out, fmt->out,
-			logouttime, length);
+			"%-8.*s%c%-12.12s%c%-*.*s%c%-*.*s%c%s\n",
+			ctl->name_len, p->ut_user, ctl->separator, utline, ctl->separator,
+			fmt->in_len, fmt->in_len, logintime, ctl->separator, fmt->out_len, fmt->out_len,
+			logouttime, ctl->separator, length);
 
 #if defined(__GLIBC__)
 #  if (__GLIBC__ == 2) && (__GLIBC_MINOR__ == 0)
@@ -529,8 +553,7 @@ static int list(const struct last_control *ctl, struct utmp *p, time_t logout_ti
 	/*
 	 *	Print out "final" string safely.
 	 */
-	for (s = final; *s; s++)
-		fputc_careful(*s, stdout, '*');
+	fputs_careful(final, stdout, '*', false, 0);
 
 	if (len < 0 || (size_t)len >= sizeof(final))
 		putchar('\n');
@@ -542,9 +565,10 @@ static int list(const struct last_control *ctl, struct utmp *p, time_t logout_ti
 	return 0;
 }
 
-
-static void __attribute__((__noreturn__)) usage(const struct last_control *ctl, FILE *out)
+#ifndef FUZZ_TARGET
+static void __attribute__((__noreturn__)) usage(const struct last_control *ctl)
 {
+	FILE *out = stdout;
 	fputs(USAGE_HEADER, out);
 	fprintf(out, _(
 		" %s [options] [<username>...] [<tty>...]\n"), program_invocation_short_name);
@@ -564,6 +588,7 @@ static void __attribute__((__noreturn__)) usage(const struct last_control *ctl,
 	fputs(_(" -R, --nohostname     don't display the hostname field\n"), out);
 	fputs(_(" -s, --since <time>   display the lines since the specified time\n"), out);
 	fputs(_(" -t, --until <time>   display the lines until the specified time\n"), out);
+	fputs(_(" -T, --tab-separated	use tabs as delimiters\n"), out);
 	fputs(_(" -p, --present <time> display who were present at the specified time\n"), out);
 	fputs(_(" -w, --fullnames      display full user and domain names\n"), out);
 	fputs(_(" -x, --system         display system shutdown entries and run level changes\n"), out);
@@ -571,25 +596,28 @@ static void __attribute__((__noreturn__)) usage(const struct last_control *ctl,
 		"                               notime|short|full|iso\n"), out);
 
 	fputs(USAGE_SEPARATOR, out);
-	fputs(USAGE_HELP, out);
-	fputs(USAGE_VERSION, out);
+	fprintf(out, USAGE_HELP_OPTIONS(22));
 	fprintf(out, USAGE_MAN_TAIL("last(1)"));
 
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
+#endif
 
-static int is_phantom(const struct last_control *ctl, struct utmp *ut)
+static int is_phantom(const struct last_control *ctl, struct utmpx *ut)
 {
 	struct passwd *pw;
-	char path[32];
+	char path[sizeof(ut->ut_line) + 16];
+	char user[sizeof(ut->ut_user) + 1];
 	int ret = 0;
 
-	if (ut->UL_UT_TIME < ctl->boot_time.tv_sec)
+	if (ut->ut_tv.tv_sec < ctl->boot_time.tv_sec)
 		return 1;
-	pw = getpwnam(ut->ut_user);
+
+	mem2strcpy(user, ut->ut_user, sizeof(ut->ut_user), sizeof(user));
+	pw = getpwnam(user);
 	if (!pw)
 		return 1;
-	sprintf(path, "/proc/%u/loginuid", ut->ut_pid);
+	snprintf(path, sizeof(path), "/proc/%u/loginuid", ut->ut_pid);
 	if (access(path, R_OK) == 0) {
 		unsigned int loginuid;
 		FILE *f = NULL;
@@ -603,8 +631,11 @@ static int is_phantom(const struct last_control *ctl, struct utmp *ut)
 			return 1;
 	} else {
 		struct stat st;
+		char utline[sizeof(ut->ut_line) + 1];
 
-		sprintf(path, "/dev/%s", ut->ut_line);
+		mem2strcpy(utline, ut->ut_line, sizeof(ut->ut_line), sizeof(utline));
+
+		snprintf(path, sizeof(path), "/dev/%s", utline);
 		if (stat(path, &st))
 			return 1;
 		if (pw->pw_uid != st.st_uid)
@@ -616,9 +647,9 @@ static int is_phantom(const struct last_control *ctl, struct utmp *ut)
 static void process_wtmp_file(const struct last_control *ctl,
 			      const char *filename)
 {
-	FILE *fp;		/* Filepointer of wtmp file */
+	FILE *fp;		/* File pointer of wtmp file */
 
-	struct utmp ut;		/* Current utmp entry */
+	struct utmpx ut;	/* Current utmp entry */
 	struct utmplist *ulist = NULL;	/* All entries */
 	struct utmplist *p;	/* Pointer into utmplist */
 	struct utmplist *next;	/* Pointer into utmplist */
@@ -634,17 +665,23 @@ static void process_wtmp_file(const struct last_control *ctl,
 	int quit = 0;		/* Flag */
 	int down = 0;		/* Down flag */
 
+#ifndef FUZZ_TARGET
 	time(&lastdown);
+#else
+	lastdown = 1596001948;
+#endif
 	/*
 	 * Fill in 'lastdate'
 	 */
 	lastdate = currentdate = lastrch = lastdown;
 
+#ifndef FUZZ_TARGET
 	/*
 	 * Install signal handlers
 	 */
 	signal(SIGINT, int_handler);
 	signal(SIGQUIT, quit_handler);
+#endif
 
 	/*
 	 * Open the utmp file
@@ -661,7 +698,7 @@ static void process_wtmp_file(const struct last_control *ctl,
 	 * Read first structure to capture the time field
 	 */
 	if (uread(fp, &ut, NULL, filename) == 1)
-		begintime = ut.UL_UT_TIME;
+		begintime = ut.ut_tv.tv_sec;
 	else {
 		if (fstat(fileno(fp), &st) != 0)
 			err(EXIT_FAILURE, _("stat of %s failed"), filename);
@@ -683,16 +720,16 @@ static void process_wtmp_file(const struct last_control *ctl,
 		if (uread(fp, &ut, &quit, filename) != 1)
 			break;
 
-		if (ctl->since && ut.UL_UT_TIME < ctl->since)
+		if (ctl->since && ut.ut_tv.tv_sec < ctl->since)
 			continue;
 
-		if (ctl->until && ctl->until < ut.UL_UT_TIME)
+		if (ctl->until && ctl->until < ut.ut_tv.tv_sec)
 			continue;
 
-		lastdate = ut.UL_UT_TIME;
+		lastdate = ut.ut_tv.tv_sec;
 
 		if (ctl->lastb) {
-			quit = list(ctl, &ut, ut.UL_UT_TIME, R_NORMAL);
+			quit = list(ctl, &ut, ut.ut_tv.tv_sec, R_NORMAL);
 			continue;
 		}
 
@@ -715,7 +752,7 @@ static void process_wtmp_file(const struct last_control *ctl,
 		else {
 			if (ut.ut_type != DEAD_PROCESS &&
 			    ut.ut_user[0] && ut.ut_line[0] &&
-			    strcmp(ut.ut_user, "LOGIN") != 0)
+			    strncmp(ut.ut_user, "LOGIN", 5) != 0)
 				ut.ut_type = USER_PROCESS;
 			/*
 			 * Even worse, applications that write ghost
@@ -728,7 +765,7 @@ static void process_wtmp_file(const struct last_control *ctl,
 			/*
 			 * Clock changes.
 			 */
-			if (strcmp(ut.ut_user, "date") == 0) {
+			if (strncmp(ut.ut_user, "date", 4) == 0) {
 				if (ut.ut_line[0] == '|')
 					ut.ut_type = OLD_TIME;
 				if (ut.ut_line[0] == '{')
@@ -742,7 +779,7 @@ static void process_wtmp_file(const struct last_control *ctl,
 				strcpy(ut.ut_line, "system down");
 				quit = list(ctl, &ut, lastboot, R_NORMAL);
 			}
-			lastdown = lastrch = ut.UL_UT_TIME;
+			lastdown = lastrch = ut.ut_tv.tv_sec;
 			down = 1;
 			break;
 		case OLD_TIME:
@@ -756,22 +793,25 @@ static void process_wtmp_file(const struct last_control *ctl,
 			break;
 		case BOOT_TIME:
 			strcpy(ut.ut_line, "system boot");
-			quit = list(ctl, &ut, lastdown, R_REBOOT);
-			lastboot = ut.UL_UT_TIME;
+			if (lastdown > lastboot && lastdown != currentdate)
+				quit = list(ctl, &ut, lastboot, R_REBOOT_CRASH);
+			else
+				quit = list(ctl, &ut, lastdown, R_REBOOT);
+			lastboot = ut.ut_tv.tv_sec;
 			down = 1;
 			break;
 		case RUN_LVL:
 			x = ut.ut_pid & 255;
 			if (ctl->extended) {
-				sprintf(ut.ut_line, "(to lvl %c)", x);
+				snprintf(ut.ut_line, sizeof(ut.ut_line), "(to lvl %c)", x);
 				quit = list(ctl, &ut, lastrch, R_NORMAL);
 			}
 			if (x == '0' || x == '6') {
-				lastdown = ut.UL_UT_TIME;
+				lastdown = ut.ut_tv.tv_sec;
 				down = 1;
 				ut.ut_type = SHUTDOWN_TIME;
 			}
-			lastrch = ut.UL_UT_TIME;
+			lastrch = ut.ut_tv.tv_sec;
 			break;
 
 		case USER_PROCESS:
@@ -784,10 +824,10 @@ static void process_wtmp_file(const struct last_control *ctl,
 			for (p = ulist; p; p = next) {
 				next = p->next;
 				if (strncmp(p->ut.ut_line, ut.ut_line,
-				    UT_LINESIZE) == 0) {
+				    sizeof(ut.ut_line)) == 0) {
 					/* Show it */
 					if (c == 0) {
-						quit = list(ctl, &ut, p->ut.UL_UT_TIME, R_NORMAL);
+						quit = list(ctl, &ut, p->ut.ut_tv.tv_sec, R_NORMAL);
 						c = 1;
 					}
 					if (p->next)
@@ -813,7 +853,7 @@ static void process_wtmp_file(const struct last_control *ctl,
 					c = whydown;
 				quit = list(ctl, &ut, lastboot, c);
 			}
-			/* FALLTHRU */
+			/* fallthrough */
 
 		case DEAD_PROCESS:
 			/*
@@ -823,7 +863,7 @@ static void process_wtmp_file(const struct last_control *ctl,
 			if (ut.ut_line[0] == 0)
 				break;
 			p = xmalloc(sizeof(struct utmplist));
-			memcpy(&p->ut, &ut, sizeof(struct utmp));
+			memcpy(&p->ut, &ut, sizeof(struct utmpx));
 			p->next  = ulist;
 			p->prev  = NULL;
 			if (ulist)
@@ -834,12 +874,14 @@ static void process_wtmp_file(const struct last_control *ctl,
 		case EMPTY:
 		case INIT_PROCESS:
 		case LOGIN_PROCESS:
+#ifdef ACCOUNTING
 		case ACCOUNTING:
+#endif
 			/* ignored ut_types */
 			break;
 
 		default:
-			warnx("unrecogized ut_type: %d", ut.ut_type);
+			warnx("unrecognized ut_type: %d", ut.ut_type);
 		}
 
 		/*
@@ -847,7 +889,7 @@ static void process_wtmp_file(const struct last_control *ctl,
 		 * the entire current ulist.
 		 */
 		if (down) {
-			lastboot = ut.UL_UT_TIME;
+			lastboot = ut.ut_tv.tv_sec;
 			whydown = (ut.ut_type == SHUTDOWN_TIME) ? R_DOWN : R_CRASH;
 			for (p = ulist; p; p = next) {
 				next = p->next;
@@ -858,7 +900,19 @@ static void process_wtmp_file(const struct last_control *ctl,
 		}
 	}
 
-	printf(_("\n%s begins %s"), basename(filename), ctime(&begintime));
+	if (ctl->time_fmt != LAST_TIMEFTM_NONE) {
+		struct last_timefmt *fmt;
+		char timestr[LAST_TIMESTAMP_LEN];
+		char *tmp = xstrdup(filename);
+
+		fmt = &timefmts[ctl->time_fmt];
+		if (time_formatter(fmt->in_fmt, timestr,
+				   sizeof(timestr), &begintime) < 0)
+			errx(EXIT_FAILURE, _("preallocation size exceeded"));
+		printf(_("\n%s begins %s\n"), basename(tmp), timestr);
+		free(tmp);
+	}
+
 	fclose(fp);
 
 	for (p = ulist; p; p = next) {
@@ -867,12 +921,43 @@ static void process_wtmp_file(const struct last_control *ctl,
 	}
 }
 
+#ifdef FUZZ_TARGET
+# include "all-io.h"
+
+int LLVMFuzzerTestOneInput(const uint8_t *data, size_t size) {
+	struct last_control ctl = {
+		.showhost = TRUE,
+		.name_len = LAST_LOGIN_LEN,
+		.time_fmt = LAST_TIMEFTM_SHORT,
+		.domain_len = LAST_DOMAIN_LEN,
+		.boot_time = {
+			.tv_sec = 1595978419,
+			.tv_usec = 816074
+		}
+	};
+	char name[] = "/tmp/test-last-fuzz.XXXXXX";
+	int fd;
+
+	fd = mkstemp_cloexec(name);
+	if (fd < 0)
+		err(EXIT_FAILURE, "mkstemp() failed");
+	if (write_all(fd, data, size) != 0)
+		err(EXIT_FAILURE, "write() failed");
+
+	process_wtmp_file(&ctl, name);
+
+	close(fd);
+	unlink(name);
+
+	return 0;
+}
+#else
 int main(int argc, char **argv)
 {
 	struct last_control ctl = {
 		.showhost = TRUE,
 		.name_len = LAST_LOGIN_LEN,
-		.time_fmt = LAST_TIMEFTM_SHORT_CTIME,
+		.time_fmt = LAST_TIMEFTM_SHORT,
 		.domain_len = LAST_DOMAIN_LEN
 	};
 	char **files = NULL;
@@ -898,10 +983,11 @@ int main(int argc, char **argv)
 	      { "ip",         no_argument,       NULL, 'i' },
 	      { "fulltimes",  no_argument,       NULL, 'F' },
 	      { "fullnames",  no_argument,       NULL, 'w' },
+	      { "tab-separated",  no_argument,   NULL, 'T' },
 	      { "time-format", required_argument, NULL, OPT_TIME_FORMAT },
 	      { NULL, 0, NULL, 0 }
 	};
-	static const ul_excl_t excl[] = {	/* rows and cols in in ASCII order */
+	static const ul_excl_t excl[] = {	/* rows and cols in ASCII order */
 		{ 'F', OPT_TIME_FORMAT },	/* fulltime, time-format */
 		{ 0 }
 	};
@@ -910,23 +996,23 @@ int main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
-	atexit(close_stdout);
+	close_stdout_atexit();
 	/*
 	 * Which file do we want to read?
 	 */
 	ctl.lastb = strcmp(program_invocation_short_name, "lastb") == 0 ? 1 : 0;
+	ctl.separator = ' ';
 	while ((c = getopt_long(argc, argv,
-			"hVf:n:RxadFit:p:s:0123456789w", long_opts, NULL)) != -1) {
+			 "hVf:n:RxadFit:p:s:T0123456789w", long_opts, NULL)) != -1) {
 
 		err_exclusive_options(c, long_opts, excl, excl_st);
 
 		switch(c) {
 		case 'h':
-			usage(&ctl, stdout);
+			usage(&ctl);
 			break;
 		case 'V':
-			printf(UTIL_LINUX_VERSION);
-			return EXIT_SUCCESS;
+			print_version(EXIT_SUCCESS);
 		case 'R':
 			ctl.showhost = 0;
 			break;
@@ -951,7 +1037,7 @@ int main(int argc, char **argv)
 			ctl.altlist = 1;
 			break;
 		case 'F':
-			ctl.time_fmt = LAST_TIMEFTM_FULL_CTIME;
+			ctl.time_fmt = LAST_TIMEFTM_CTIME;
 			break;
 		case 'p':
 			if (parse_timestamp(optarg, &p) < 0)
@@ -969,10 +1055,10 @@ int main(int argc, char **argv)
 			ctl.until = (time_t) (p / 1000000);
 			break;
 		case 'w':
-			if (ctl.name_len < UT_NAMESIZE)
-				ctl.name_len = UT_NAMESIZE;
-			if (ctl.domain_len < UT_HOSTSIZE)
-				ctl.domain_len = UT_HOSTSIZE;
+			if (ctl.name_len < sizeof_member(struct utmpx, ut_user))
+				ctl.name_len = sizeof_member(struct utmpx, ut_user);
+			if (ctl.domain_len < sizeof_member(struct utmpx, ut_host))
+				ctl.domain_len = sizeof_member(struct utmpx, ut_host);
 			break;
 		case '0': case '1': case '2': case '3': case '4':
 		case '5': case '6': case '7': case '8': case '9':
@@ -981,9 +1067,11 @@ int main(int argc, char **argv)
 		case OPT_TIME_FORMAT:
 			ctl.time_fmt = which_time_format(optarg);
 			break;
-		default:
-			usage(&ctl, stderr);
+		case 'T':
+			ctl.separator = '\t';
 			break;
+		default:
+			errtryhelp(EXIT_FAILURE);
 		}
 	}
 
@@ -1003,3 +1091,4 @@ int main(int argc, char **argv)
 	free(files);
 	return EXIT_SUCCESS;
 }
+#endif
